import re
from collections.abc import Iterator, Sequence
from decimal import Decimal
from functools import cached_property
from typing import TYPE_CHECKING, Any, ClassVar, Optional, Union, cast

import rlp  # type: ignore
from cchecksum import to_checksum_address
from eth_abi import decode, encode
from eth_abi.exceptions import InsufficientDataBytes, NonEmptyPaddingBytes
from eth_pydantic_types import HexBytes
from eth_typing import Hash32, HexStr
from eth_utils import (
    add_0x_prefix,
    encode_hex,
    humanize_hash,
    is_0x_prefixed,
    is_hex,
    is_hex_address,
    keccak,
    to_bytes,
    to_hex,
)
from ethpm_types.abi import ABIType, ConstructorABI, EventABI, MethodABI
from pydantic import Field, computed_field, field_validator, model_validator
from pydantic_settings import SettingsConfigDict

from ape.api.config import PluginConfig
from ape.api.networks import EcosystemAPI
from ape.api.providers import BlockAPI
from ape.contracts.base import ContractCall
from ape.exceptions import (
    ApeException,
    APINotImplementedError,
    ConversionError,
    CustomError,
    DecodingError,
    SignatureError,
)
from ape.logging import logger
from ape.managers.config import merge_configs
from ape.types.address import AddressType, RawAddress
from ape.types.basic import HexInt
from ape.types.events import ContractLog
from ape.types.gas import AutoGasLimit, GasLimit
from ape.types.signatures import TransactionSignature
from ape.types.units import CurrencyValueComparable
from ape.utils.abi import LogInputABICollection, Struct, StructParser, is_array, returns_array
from ape.utils.basemodel import _assert_not_ipython_check, only_raise_attribute_error
from ape.utils.misc import (
    DEFAULT_LIVE_NETWORK_BASE_FEE_MULTIPLIER,
    DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT,
    DEFAULT_MAX_RETRIES_TX,
    DEFAULT_TRANSACTION_ACCEPTANCE_TIMEOUT,
    DEFAULT_TRANSACTION_TYPE,
    EMPTY_BYTES32,
    LOCAL_NETWORK_NAME,
    ZERO_ADDRESS,
)
from ape_ethereum.proxies import (
    IMPLEMENTATION_ABI,
    MASTER_COPY_ABI,
    PROXY_TYPE_ABI,
    ProxyInfo,
    ProxyType,
)
from ape_ethereum.trace import _REVERT_PREFIX, Trace, TransactionTrace
from ape_ethereum.transactions import (
    AccessListTransaction,
    BaseTransaction,
    DynamicFeeTransaction,
    Receipt,
    SetCodeTransaction,
    SharedBlobReceipt,
    SharedBlobTransaction,
    StaticFeeTransaction,
    TransactionStatusEnum,
    TransactionType,
)

if TYPE_CHECKING:
    from ethpm_types import ContractType

    from ape.api.trace import TraceAPI
    from ape.api.transactions import ReceiptAPI, TransactionAPI


NETWORKS = {
    # chain_id, network_id
    "mainnet": (1, 1),
    "sepolia": (11155111, 11155111),
    "holesky": (17000, 17000),
}
BLUEPRINT_HEADER = HexBytes("0xfe71")


class NetworkConfig(PluginConfig):
    """
    The Ethereum network config base class for each
    network, e.g. ``"mainnet"``, ```"local"``, etc.
    """

    required_confirmations: int = 0
    """
    The amount of blocks to wait before
    considering a transaction 'confirmed'.
    """

    default_provider: Optional[str] = "node"
    """
    The default provider to use. If set to ``None``, ape will rely on
    an external plugin supplying the provider implementation, such as
    ``ape-hardhat`` supplying forked-network providers.
    """

    block_time: int = 0
    """
    Approximate amount of time for a block to be
    added to the network.
    """

    transaction_acceptance_timeout: int = DEFAULT_TRANSACTION_ACCEPTANCE_TIMEOUT
    """
    The amount tof time before failing when sending a
    transaction and it leaving the mempool.
    """

    default_transaction_type: TransactionType = TransactionType.DYNAMIC
    """
    The default type of transaction to use.
    """

    max_receipt_retries: int = DEFAULT_MAX_RETRIES_TX
    """
    Maximum number of retries when getting a receipt
    from a transaction before failing.
    """

    gas_limit: GasLimit = "auto"
    """
    The gas limit override to use for the network. If set to ``"auto"``, ape will
    estimate gas limits based on the transaction. If set to ``"max"`` the gas limit
    will be set to the maximum block gas limit for the network. Otherwise an ``int``
    can be used to specify an explicit gas limit amount (either base 10 or 16).

    The default for local networks is ``"max"``, otherwise ``"auto"``.
    """

    base_fee_multiplier: float = 1.0
    """A multiplier to apply to a transaction base fee."""

    is_mainnet: Optional[bool] = None
    """
    Set to ``True`` to declare as a mainnet or ``False`` to ensure
    it isn't detected as one.
    """

    request_headers: dict = {}
    """Optionally config extra request headers whenever using this network."""

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_ETHEREUM_")

    @field_validator("gas_limit", mode="before")
    @classmethod
    def validate_gas_limit(cls, value):
        if isinstance(value, dict) and "auto" in value:
            return AutoGasLimit.model_validate(value["auto"])

        elif value in ("auto", "max") or isinstance(value, AutoGasLimit):
            return value

        elif isinstance(value, int):
            return value

        elif isinstance(value, str) and value.isnumeric():
            return int(value)

        elif isinstance(value, str) and is_hex(value) and is_0x_prefixed(value):
            return int(value, 16)

        elif is_hex(value):
            raise ValueError("Gas limit hex str must include '0x' prefix.")

        raise ValueError(f"Invalid gas limit '{value}'")


class ForkedNetworkConfig(NetworkConfig):
    upstream_provider: Optional[str] = None
    """
    The provider to use as the upstream-provider for this forked network.
    """


def create_local_network_config(
    default_provider: Optional[str] = None, use_fork: bool = False, **kwargs
):
    if "gas_limit" not in kwargs:
        kwargs["gas_limit"] = "max"

    return create_network_config(
        base_fee_multiplier=1.0,
        default_provider=default_provider,
        required_confirmations=0,
        transaction_acceptance_timeout=DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT,
        cls=ForkedNetworkConfig if use_fork else NetworkConfig,
        **kwargs,
    )


def create_network_config(
    required_confirmations: int = 2,
    base_fee_multiplier: float = DEFAULT_LIVE_NETWORK_BASE_FEE_MULTIPLIER,
    cls: type = NetworkConfig,
    **kwargs,
) -> NetworkConfig:
    return cls(
        base_fee_multiplier=base_fee_multiplier,
        required_confirmations=required_confirmations,
        **kwargs,
    )


class BaseEthereumConfig(PluginConfig):
    """
    L2 plugins should use this as their config base-class.
    """

    DEFAULT_TRANSACTION_TYPE: ClassVar[int] = TransactionType.DYNAMIC.value
    DEFAULT_LOCAL_GAS_LIMIT: ClassVar[GasLimit] = "max"
    NETWORKS: ClassVar[dict[str, tuple[int, int]]] = NETWORKS

    default_network: str = LOCAL_NETWORK_NAME
    _forked_configs: dict[str, ForkedNetworkConfig] = {}
    _custom_networks: dict[str, NetworkConfig] = {}

    # NOTE: This gets appended to Ape's root User-Agent string.
    request_headers: dict = {}

    model_config = SettingsConfigDict(extra="allow", env_prefix="APE_ETHEREUM_")

    @model_validator(mode="before")
    @classmethod
    def load_network_configs(cls, values):
        cfg_forks: dict[str, ForkedNetworkConfig] = {}
        custom_networks = {}
        for name, obj in values.items():
            if name.startswith("_"):
                continue

            net_name = name.replace("-", "_")
            key = net_name.replace("_fork", "")
            if net_name.endswith("_fork"):
                key = net_name.replace("_fork", "")
                default_fork_model = create_local_network_config(
                    use_fork=True,
                    default_transaction_type=cls.DEFAULT_TRANSACTION_TYPE,
                    gas_limit=cls.DEFAULT_LOCAL_GAS_LIMIT,
                ).model_dump(by_alias=True)
                data = merge_configs(default_fork_model, obj)
                cfg_forks[key] = ForkedNetworkConfig.model_validate(data)

            elif (
                key != LOCAL_NETWORK_NAME
                and key not in cls.NETWORKS
                and isinstance(obj, dict)
                and key not in ("request_headers",)
            ):
                # Custom network.
                default_network_model = create_network_config(
                    default_transaction_type=cls.DEFAULT_TRANSACTION_TYPE
                ).model_dump(by_alias=True)
                data = merge_configs(default_network_model, obj)
                custom_networks[name] = NetworkConfig.model_validate(data)

        values["_forked_configs"] = {**cfg_forks, **values.get("_forked_configs", {})}
        return {**values, **custom_networks}

    @computed_field  # type: ignore[misc]
    @cached_property
    def local(self) -> NetworkConfig:
        return create_local_network_config(
            default_provider="test",
            default_transaction_type=self.DEFAULT_TRANSACTION_TYPE,
            gas_limit=self.DEFAULT_LOCAL_GAS_LIMIT,
        )

    @only_raise_attribute_error
    def __getattr__(self, key: str) -> Any:
        _assert_not_ipython_check(key)
        net_key = key.replace("-", "_")
        if net_key.endswith("_fork"):
            return self._get_forked_config(net_key)

        try:
            return super().__getattr__(key)
        except AttributeError:
            return NetworkConfig(default_transaction_type=self.DEFAULT_TRANSACTION_TYPE)

    def __contains__(self, key: str) -> bool:
        net_key = key.replace("-", "_")
        if net_key.endswith("_fork"):
            return self._get_forked_config(net_key) is not None

        return super().__contains__(key)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        net_key = key.replace("-", "_")
        if net_key.endswith("_fork"):
            if cfg := self._get_forked_config(net_key):
                return cfg

        result: Any
        if result := super().get(key, default=default):
            return result

        # Handle weird base-class differences.
        try:
            return self.__getattr__(key)
        except AttributeError:
            return default

    def _get_forked_config(self, name: str) -> Optional[ForkedNetworkConfig]:
        live_key: str = name.replace("_fork", "")
        if self._forked_configs.get(live_key):
            return self._forked_configs[live_key]

        live_cfg: Any
        if live_cfg := self.get(live_key):
            if isinstance(live_cfg, NetworkConfig):
                fork_cfg = create_local_network_config(
                    use_fork=True,
                    default_transaction_type=self.DEFAULT_TRANSACTION_TYPE,
                    gas_limit=self.DEFAULT_LOCAL_GAS_LIMIT,
                )
                self._forked_configs[live_key] = fork_cfg
                return fork_cfg

        return None

    def _get_custom_network(self, name: str) -> NetworkConfig:
        return self._custom_networks.get(name, NetworkConfig())


class EthereumConfig(BaseEthereumConfig):
    mainnet: NetworkConfig = create_network_config(block_time=13)
    holesky: NetworkConfig = create_network_config(block_time=13)
    sepolia: NetworkConfig = create_network_config(block_time=15)


class Block(BlockAPI):
    """
    Class for representing a block on a chain.
    """

    gas_limit: HexInt = Field(alias="gasLimit")
    gas_used: HexInt = Field(alias="gasUsed")
    base_fee: HexInt = Field(default=0, alias="baseFeePerGas")
    difficulty: HexInt = 0
    total_difficulty: HexInt = Field(default=0, alias="totalDifficulty")
    uncles: list[HexBytes] = []

    # Type re-declares.
    hash: Optional[HexBytes] = None
    parent_hash: HexBytes = Field(
        default=EMPTY_BYTES32, alias="parentHash"
    )  # NOTE: genesis block has no parent hash

    @computed_field()  # type: ignore[misc]
    @property
    def size(self) -> int:
        if self._size is not None:
            # The size was provided with the rest of the model
            # (normal).
            return self._size

        number = self.number
        if number is None:
            raise APINotImplementedError()

        # Try to get it from the provider.
        elif provider := self.network_manager.active_provider:
            block = provider.get_block(number)
            size = block._size
            if size is not None and size > -1:
                self._size = size
                return size

        raise APINotImplementedError()


class Ethereum(EcosystemAPI):
    # NOTE: `default_transaction_type` should be overridden
    #   if the chain doesn't support EIP-1559.

    fee_token_symbol: str = "ETH"

    @property
    def config(self) -> EthereumConfig:
        return cast(EthereumConfig, super().config)

    @property
    def default_transaction_type(self) -> TransactionType:
        if provider := self.network_manager.active_provider:
            # Check connected network first.
            networks_to_check = [provider.network.name, self.default_network_name]
        else:
            networks_to_check = [self.default_network_name]

        for name in networks_to_check:
            network = self.get_network(name)
            ecosystem_config = network.ecosystem_config
            ecosystem_default = ecosystem_config.get(
                "default_transaction_type", DEFAULT_TRANSACTION_TYPE
            )
            result: int = network.config.get("default_transaction_type", ecosystem_default)
            return TransactionType(result)

        return TransactionType(DEFAULT_TRANSACTION_TYPE)

    @classmethod
    def decode_address(cls, raw_address: RawAddress) -> AddressType:
        return to_checksum_address(HexBytes(raw_address)[-20:].rjust(20, b"\x00"))

    @classmethod
    def encode_address(cls, address: AddressType) -> RawAddress:
        return f"{address}"

    def decode_transaction_type(self, transaction_type_id: Any) -> type["TransactionAPI"]:
        if isinstance(transaction_type_id, TransactionType):
            tx_type = transaction_type_id
        elif isinstance(transaction_type_id, int):
            tx_type = TransactionType(transaction_type_id)
        else:
            # Using hex or alike.
            tx_type = self.conversion_manager.convert(transaction_type_id, int)

        if tx_type is TransactionType.STATIC:
            return StaticFeeTransaction

        elif tx_type is TransactionType.ACCESS_LIST:
            return AccessListTransaction

        elif tx_type is TransactionType.SET_CODE:
            return SetCodeTransaction

        return DynamicFeeTransaction

    def encode_contract_blueprint(
        self, contract_type: "ContractType", *args, **kwargs
    ) -> "TransactionAPI":
        # EIP-5202 implementation.
        bytes_obj = contract_type.deployment_bytecode
        contract_bytes = (bytes_obj.to_bytes() or b"") if bytes_obj else b""
        header = kwargs.pop("header", BLUEPRINT_HEADER)
        blueprint_bytecode = header + b"\x00" + contract_bytes
        len_bytes = len(blueprint_bytecode).to_bytes(2, "big")
        return_data_size = kwargs.pop("return_data_size", HexBytes("0x61"))
        return_instructions = kwargs.pop("return_instructions", HexBytes("0x3d81600a3d39f3"))
        deploy_bytecode = HexBytes(
            return_data_size + len_bytes + return_instructions + blueprint_bytecode
        )
        converted_kwargs = self.conversion_manager.convert_method_kwargs(kwargs)
        return self.encode_deployment(
            deploy_bytecode, contract_type.constructor, **converted_kwargs
        )

    def get_proxy_info(self, address: AddressType) -> Optional[ProxyInfo]:
        contract_code = self.chain_manager.get_code(address)
        if isinstance(contract_code, bytes):
            contract_code = to_hex(contract_code)

        if not (code := contract_code[2:]):
            return None

        patterns = {
            ProxyType.Minimal: r"^363d3d373d3d3d363d73(.{40})5af43d82803e903d91602b57fd5bf3",
            ProxyType.ZeroAge: r"^3d3d3d3d363d3d37363d73(.{40})5af43d3d93803e602a57fd5bf3",
            ProxyType.Clones: r"^36603057343d52307f830d2d700a97af574b186c80d40429385d24241565b08a7c559ba283a964d9b160203da23d3df35b3d3d3d3d363d3d37363d73(.{40})5af43d3d93803e605b57fd5bf3",
            ProxyType.Vyper: r"^366000600037611000600036600073(.{40})5af4602c57600080fd5b6110006000f3",
            ProxyType.VyperBeta: r"^366000600037611000600036600073(.{40})5af41558576110006000f3",
            ProxyType.CWIA: r"^3d3d3d3d363d3d3761.{4}603736393661.{4}013d73(.{40})5af43d3d93803e603557fd5bf3.*",
            ProxyType.OldCWIA: r"^363d3d3761.{4}603836393d3d3d3661.{4}013d73(.{40})5af43d82803e903d91603657fd5bf3.*",
            ProxyType.SudoswapCWIA: r"^3d3d3d3d363d3d37605160353639366051013d73(.{40})5af43d3d93803e603357fd5bf3.*",
            ProxyType.SoladyCWIA: r"36602c57343d527f9e4ac34f21c619cefc926c8bd93b54bf5a39c7ab2127a895af1cc0691d7e3dff593da1005b363d3d373d3d3d3d61.{4}806062363936013d73(.{40})5af43d3d93803e606057fd5bf3.*",
            ProxyType.SplitsCWIA: r"36602f57343d527f9e4ac34f21c619cefc926c8bd93b54bf5a39c7ab2127a895af1cc0691d7e3dff60203da13d3df35b3d3d3d3d363d3d3761.{4}606736393661.{4}013d73(.{40})5af43d3d93803e606557fd5bf3.*",
            ProxyType.SoladyPush0: r"^5f5f365f5f37365f73(.{40})5af43d5f5f3e6029573d5ffd5b3d5ff3",
            ProxyType.SetCode: r"^ef0100(.{40})$",
        }
        for type_, pattern in patterns.items():
            if match := re.match(pattern, code):
                target = self.conversion_manager.convert(match.group(1), AddressType)
                return ProxyInfo(type=type_, target=target)

        sequence_pattern = r"363d3d373d3d3d363d30545af43d82803e903d91601857fd5bf3"
        if re.match(sequence_pattern, code):
            # the implementation is stored in the slot matching proxy address
            slot = self.provider.get_storage(address, address)
            target = self.conversion_manager.convert(slot[-20:], AddressType)
            return ProxyInfo(type=ProxyType.Sequence, target=target)

        def str_to_slot(text):
            return int(to_hex(keccak(text=text)), 16)

        slots = {
            ProxyType.Standard: str_to_slot("eip1967.proxy.implementation") - 1,
            ProxyType.Beacon: str_to_slot("eip1967.proxy.beacon") - 1,
            ProxyType.OpenZeppelin: str_to_slot("org.zeppelinos.proxy.implementation"),
            ProxyType.UUPS: str_to_slot("PROXIABLE"),
        }
        for _type, slot in slots.items():
            try:
                # TODO perf: use a batch call here when ape adds support
                storage = self.provider.get_storage(address, slot)
            except NotImplementedError:
                # Break early on not-implemented error rather than attempting
                # to try more proxy types.
                break

            if sum(storage) == 0:
                continue

            target = self.conversion_manager.convert(storage[-20:], AddressType)
            # read `target.implementation()`
            if _type == ProxyType.Beacon:
                target = ContractCall(IMPLEMENTATION_ABI, target)(skip_trace=True)

            return ProxyInfo(type=_type, target=target, abi=IMPLEMENTATION_ABI)

        # safe >=1.1.0 provides `masterCopy()`, which is also stored in slot 0
        # call it and check that target matches
        try:
            singleton = ContractCall(MASTER_COPY_ABI, address)(skip_trace=True)
            slot_0 = self.provider.get_storage(address, 0)
            target = self.conversion_manager.convert(slot_0[-20:], AddressType)
            # NOTE: `target` is set in initialized proxies
            if target != ZERO_ADDRESS and target == singleton:
                return ProxyInfo(type=ProxyType.GnosisSafe, target=target, abi=MASTER_COPY_ABI)

        except ApeException:
            pass

        # eip-897 delegate proxy, read `proxyType()` and `implementation()`
        # perf: only make a call when a proxyType() selector is mentioned in the code
        eip897_pattern = b"\x63" + keccak(text="proxyType()")[:4]
        if eip897_pattern.hex() in code:
            try:
                proxy_type = ContractCall(PROXY_TYPE_ABI, address)(skip_trace=True)
                if proxy_type not in (1, 2):
                    raise ValueError(f"ProxyType '{proxy_type}' not permitted by EIP-897.")

                target = ContractCall(IMPLEMENTATION_ABI, address)(skip_trace=True)
                # avoid recursion
                if target != ZERO_ADDRESS:
                    return ProxyInfo(type=ProxyType.Delegate, target=target, abi=IMPLEMENTATION_ABI)

            except (ApeException, ValueError):
                pass

        return None

    def decode_receipt(self, data: dict) -> "ReceiptAPI":
        status = data.get("status")
        if status is not None:
            status = self.conversion_manager.convert(status, int)
            status = TransactionStatusEnum(status)

        hash_key_choices = (
            "hash",
            "txHash",
            "txn_hash",
            "txnHash",
            "transactionHash",
            "transaction_hash",
        )
        txn_hash = next((data[choice] for choice in hash_key_choices if choice in data), None)
        if txn_hash and isinstance(txn_hash, bytes):
            txn_hash = to_hex(txn_hash)

        data_bytes = data.get("data")
        if data_bytes and isinstance(data_bytes, str):
            data["data"] = HexBytes(data_bytes)
        elif "input" in data and isinstance(data["input"], str):
            data["input"] = HexBytes(data["input"])

        block_number = data.get("block_number", data.get("blockNumber"))
        if block_number is None:
            raise ValueError("Missing block number.")

        receipt_kwargs = {
            "block_number": block_number,
            "contract_address": data.get("contract_address", data.get("contractAddress")),
            "gas_limit": data.get("gas", data.get("gas_limit", data.get("gasLimit"))) or 0,
            "gas_price": data.get("gas_price", data.get("gasPrice")) or 0,
            "gas_used": data.get("gas_used", data.get("gasUsed")) or 0,
            "logs": data.get("logs", []),
            "status": status,
            "txn_hash": txn_hash,
            "transaction": self.create_transaction(**data),
        }

        receipt_cls: type[Receipt]
        if data.get("type") == 3:
            receipt_cls = SharedBlobReceipt
            blob_gas_price = data.get("blob_gas_price")
            if blob_gas_price is None:
                blob_gas_price = data.get("blobGasPrice")

            receipt_kwargs["blobGasPrice"] = blob_gas_price
            receipt_kwargs["blobGasUsed"] = data.get("blob_gas_used", data.get("blobGasUsed")) or 0

        else:
            receipt_cls = Receipt

        error = receipt_kwargs.pop("error", None)
        receipt = receipt_cls.model_validate(receipt_kwargs)
        receipt.error = error
        return receipt

    def decode_block(self, data: dict) -> BlockAPI:
        data["hash"] = HexBytes(data["hash"]) if data.get("hash") else None
        if "gas_limit" in data:
            data["gasLimit"] = data.pop("gas_limit")
        if "gas_used" in data:
            data["gasUsed"] = data.pop("gas_used")
        if "parent_hash" in data:
            data["parentHash"] = HexBytes(data.pop("parent_hash"))
        if "transaction_ids" in data:
            data["transactions"] = data.pop("transaction_ids")
        if "total_difficulty" in data:
            data["totalDifficulty"] = data.pop("total_difficulty") or 0
        elif "totalDifficulty" in data:
            data["totalDifficulty"] = data.pop("totalDifficulty") or 0
        if "base_fee" in data:
            data["baseFeePerGas"] = data.pop("base_fee")
        elif "baseFee" in data:
            data["baseFeePerGas"] = data.pop("baseFee")
        if "transactions" in data:
            data["num_transactions"] = len(data["transactions"])

        return Block.model_validate(data)

    def _python_type_for_abi_type(self, abi_type: ABIType) -> Union[type, Sequence]:
        # NOTE: An array can be an array of tuples, so we start with an array check
        if str(abi_type.type).endswith("]"):
            # remove one layer of the potential onion of array
            abi_type_str = str(abi_type.type)
            last_bracket_pos = abi_type_str.rfind("[")
            new_type = abi_type_str[:last_bracket_pos] if last_bracket_pos != -1 else abi_type_str

            # create a new type with the inner type of array
            new_abi_type = ABIType(type=new_type, **abi_type.model_dump(exclude={"type"}))
            # NOTE: type for static and dynamic array is a single item list
            # containing the type of the array
            return [self._python_type_for_abi_type(new_abi_type)]

        if abi_type.components is not None:
            return tuple(self._python_type_for_abi_type(c) for c in abi_type.components)

        if abi_type.type == "address":
            return AddressType

        elif abi_type.type == "bool":
            return bool

        elif abi_type.type == "string":
            return str

        elif "bytes" in abi_type.type:
            return bytes

        elif "int" in abi_type.type:
            return int

        elif "fixed" in abi_type.type:
            return Decimal

        raise ConversionError(f"Unable to convert '{abi_type}'.")

    def encode_calldata(self, abi: Union[ConstructorABI, MethodABI], *args) -> HexBytes:
        if not abi.inputs:
            return HexBytes("")

        parser = StructParser(abi)
        arguments = parser.encode_input(args)
        input_types = [i.canonical_type for i in abi.inputs]
        python_types = tuple(self._python_type_for_abi_type(i) for i in abi.inputs)
        converted_args = self.conversion_manager.convert(arguments, python_types)
        encoded_calldata = encode(input_types, converted_args)
        return HexBytes(encoded_calldata)

    def decode_calldata(self, abi: Union[ConstructorABI, MethodABI], calldata: bytes) -> dict:
        raw_input_types = [i.canonical_type for i in abi.inputs]
        input_types = [parse_type(i.model_dump()) for i in abi.inputs]

        try:
            raw_input_values = decode(raw_input_types, calldata, strict=False)
        except (InsufficientDataBytes, OverflowError, NonEmptyPaddingBytes) as err:
            raise DecodingError(str(err)) from err

        input_values = [
            self.decode_primitive_value(v, t) for v, t in zip(raw_input_values, input_types)
        ]
        arguments = {}
        index = 0
        for i, v in zip(abi.inputs, input_values):
            name = i.name or f"{index}"
            arguments[name] = v
            index += 1

        return arguments

    def decode_returndata(self, abi: MethodABI, raw_data: bytes) -> tuple[Any, ...]:
        output_types_str_ls = [o.canonical_type for o in abi.outputs]

        if raw_data:
            try:
                vm_return_values = decode(output_types_str_ls, raw_data, strict=False)
            except (InsufficientDataBytes, NonEmptyPaddingBytes) as err:
                raise DecodingError(str(err)) from err
        else:
            # Use all zeroes.
            vm_return_values = tuple([0 for _ in output_types_str_ls])

        if not vm_return_values:
            return vm_return_values

        elif not isinstance(vm_return_values, (tuple, list)):
            vm_return_values = (vm_return_values,)

        output_types = [parse_type(o.model_dump()) for o in abi.outputs]
        output_values = [
            self.decode_primitive_value(v, t) for v, t in zip(vm_return_values, output_types)
        ]
        parser = StructParser(abi)
        output_values = parser.decode_output(output_values)

        if issubclass(type(output_values), Struct):
            return (output_values,)

        elif (
            returns_array(abi)
            and isinstance(output_values, (list, tuple))
            and len(output_values) == 1
        ):
            # Array of structs or tuples: don't convert to list
            # Array of anything else: convert to single list

            if issubclass(type(output_values[0]), Struct):
                return ([output_values[0]],)

            else:
                try:
                    return ([o for o in output_values[0]],)  # type: ignore[union-attr]
                except Exception:
                    # On-chains transaction data errors.
                    return (output_values,)

        elif returns_array(abi):
            # Tuple with single item as the array.
            return (output_values,)

        return tuple(output_values)

    def _enrich_value(self, value: Any, **kwargs) -> Any:
        if isinstance(value, bytes):
            try:
                string_value = value.strip(b"\x00").decode("utf8")
                return f'"{string_value}"'
            except UnicodeDecodeError:
                # Truncate bytes if very long.
                if len(value) > 24:
                    return f"{add_0x_prefix(HexStr(humanize_hash(cast(Hash32, value))))}"

                hex_str = to_hex(value)
                if is_hex_address(hex_str):
                    return self._enrich_value(hex_str, **kwargs)

                return hex_str

        elif isinstance(value, str) and is_hex_address(value):
            address = self.decode_address(value)
            return self._enrich_contract_id(address, **kwargs)

        elif isinstance(value, str):
            # Surround non-address strings with quotes.
            return f'"{value}"'

        elif isinstance(value, int):
            return int(value)  # Eliminate int-base classes.

        elif isinstance(value, (list, tuple)):
            return [self._enrich_value(v, **kwargs) for v in value]

        elif isinstance(value, Struct):
            return {k: self._enrich_value(v, **kwargs) for k, v in value.items()}

        return value

    def decode_primitive_value(
        self, value: Any, output_type: Union[str, tuple, list]
    ) -> Union[str, HexBytes, int, tuple, list]:
        if output_type == "address":
            try:
                return self.decode_address(value)
            except InsufficientDataBytes as err:
                raise DecodingError() from err

        elif isinstance(value, bytes):
            return HexBytes(value)

        elif isinstance(value, int) and not isinstance(value, bool):
            # Wrap integers in a special type that allows us to compare
            # them with currency-value strings.
            return CurrencyValueComparable(value)

        elif isinstance(output_type, str) and is_array(output_type):
            sub_type = "[".join(output_type.split("[")[:-1])

            if not isinstance(value, (list, tuple)):
                value = (value,)

            return [self.decode_primitive_value(v, sub_type) for v in value]

        elif isinstance(output_type, tuple):
            return tuple([self.decode_primitive_value(v, t) for v, t in zip(value, output_type)])

        elif (
            isinstance(output_type, list)
            and len(output_type) == 1
            and isinstance(value, (list, tuple))
        ):
            return tuple([self.decode_primitive_value(v, output_type[0]) for v in value])

        return value

    def encode_deployment(
        self, deployment_bytecode: HexBytes, abi: ConstructorABI, *args, **kwargs
    ) -> BaseTransaction:
        txn = self.create_transaction(**kwargs)
        data = HexBytes(deployment_bytecode)

        # Encode args, if there are any
        if abi and args:
            data = HexBytes(data + self.encode_calldata(abi, *args))

        txn.data = data
        return cast(BaseTransaction, txn)

    def encode_transaction(
        self,
        address: AddressType,
        abi: MethodABI,
        *args,
        **kwargs,
    ) -> BaseTransaction:
        txn = self.create_transaction(receiver=address, **kwargs)

        # Add method ID
        txn.data = self.get_method_selector(abi)
        txn.data = HexBytes(txn.data + self.encode_calldata(abi, *args))

        return cast(BaseTransaction, txn)

    def create_transaction(self, **kwargs) -> "TransactionAPI":
        """
        Returns a transaction using the given constructor kwargs.

        **NOTE**: This generally should not be called by the user since this API method is used as a
        hook for Ecosystems to customize how transactions are created.

        Returns:
            :class:`~ape.api.transactions.TransactionAPI`
        """
        # Handle all aliases.
        tx_data = dict(kwargs)
        tx_data = _correct_key(
            "max_priority_fee",
            tx_data,
            ("max_priority_fee_per_gas", "maxPriorityFeePerGas", "maxPriorityFee"),
        )
        tx_data = _correct_key("max_fee", tx_data, ("max_fee_per_gas", "maxFeePerGas", "maxFee"))
        tx_data = _correct_key("gas", tx_data, ("gas_limit", "gasLimit"))
        tx_data = _correct_key("gas_price", tx_data, ("gasPrice",))
        tx_data = _correct_key(
            "type",
            tx_data,
            ("txType", "tx_type", "txnType", "txn_type", "transactionType", "transaction_type"),
        )
        tx_data = _correct_key("maxFeePerBlobGas", tx_data, ("max_fee_per_blob_gas",))
        tx_data = _correct_key("blobVersionedHashes", tx_data, ("blob_versioned_hashes",))

        # Handle unique value specifications, such as "1 ether".
        if "value" in tx_data and not isinstance(tx_data["value"], int):
            value = tx_data["value"] or 0  # Convert None to 0.
            tx_data["value"] = self.conversion_manager.convert(value, int)

        # None is not allowed, the user likely means `b""`.
        if "data" in tx_data and tx_data["data"] is None:
            tx_data["data"] = b""

        # Deduce the transaction type.
        transaction_types: dict[TransactionType, type[TransactionAPI]] = {
            TransactionType.STATIC: StaticFeeTransaction,
            TransactionType.ACCESS_LIST: AccessListTransaction,
            TransactionType.DYNAMIC: DynamicFeeTransaction,
            TransactionType.SHARED_BLOB: SharedBlobTransaction,
            TransactionType.SET_CODE: SetCodeTransaction,
        }
        if "type" in tx_data:
            # May be None in data.
            if tx_data["type"] is None:
                # Explicit `None` means used default.
                version = self.default_transaction_type
            elif isinstance(tx_data["type"], TransactionType):
                version = tx_data["type"]
            elif isinstance(tx_data["type"], int):
                version = TransactionType(tx_data["type"])
            else:
                # Using hex values or alike.
                version = TransactionType(self.conversion_manager.convert(tx_data["type"], int))

        # NOTE: Determine these in reverse order
        elif "authorizationList" in tx_data:
            version = TransactionType.SET_CODE
        elif "maxFeePerBlobGas" in tx_data or "blobVersionedHashes" in tx_data:
            version = TransactionType.SHARED_BLOB
        elif "max_fee" in tx_data or "max_priority_fee" in tx_data:
            version = TransactionType.DYNAMIC
        elif "access_list" in tx_data or "accessList" in tx_data:
            version = TransactionType.ACCESS_LIST
        elif "gas_price" in tx_data:
            version = TransactionType.STATIC
        else:
            version = self.default_transaction_type

        tx_data["type"] = version.value

        # This causes problems in pydantic for some reason.
        # NOTE: This must happen after deducing the tx type!
        if "gas_price" in tx_data and tx_data["gas_price"] is None:
            del tx_data["gas_price"]

        txn_class = transaction_types[version]

        if "required_confirmations" not in tx_data or tx_data["required_confirmations"] is None:
            # Attempt to use default required-confirmations from `ape-config.yaml`.
            required_confirmations = 0
            active_provider = self.network_manager.active_provider
            if active_provider:
                required_confirmations = active_provider.network.required_confirmations

            tx_data["required_confirmations"] = required_confirmations

        if isinstance(tx_data.get("chainId"), str):
            tx_data["chainId"] = int(tx_data["chainId"], 16)

        elif (
            "chainId" not in tx_data or tx_data["chainId"] is None
        ) and self.network_manager.active_provider is not None:
            tx_data["chainId"] = self.provider.chain_id

        if "input" in tx_data:
            tx_data["data"] = tx_data.pop("input")

        if all(field in tx_data for field in ("v", "r", "s")):
            tx_data["signature"] = TransactionSignature(
                v=tx_data["v"],
                r=bytes(tx_data["r"]),
                s=bytes(tx_data["s"]),
            )

        if "gas" not in tx_data:
            tx_data["gas"] = None

        return txn_class.model_validate(tx_data)

    def decode_logs(self, logs: Sequence[dict], *events: EventABI) -> Iterator[ContractLog]:
        if not logs:
            return

        abi_inputs = {
            encode_hex(keccak(text=abi.selector)): LogInputABICollection(abi) for abi in events
        }

        def get_abi(_topic: HexStr) -> Optional[LogInputABICollection]:
            return abi_inputs[_topic] if _topic in abi_inputs else None

        for log in logs:
            if log.get("anonymous"):
                raise NotImplementedError(
                    "decoding anonymous logs is not supported with this method"
                )
            topics = log["topics"]
            # web3.py converts topics to HexBytes, data is always a HexStr
            if isinstance(log["topics"][0], bytes):
                topics = [encode_hex(t) for t in log["topics"]]

            elif not topics:
                continue

            if not (abi := get_abi(topics[0])):
                continue

            event_arguments = abi.decode(topics, log["data"], use_hex_on_fail=True)

            # Since LogABICollection does not have access to the Ecosystem,
            # the rest of the decoding must happen here.
            converted_arguments: dict = {}

            for item in abi.abi.inputs:
                _type, key, value = item.canonical_type, item.name, event_arguments[item.name]

                if isinstance(value, Struct):
                    struct_types = _type.lstrip("(").rstrip(")").split(",")
                    for struct_type, (struct_key, struct_val) in zip(struct_types, value.items()):
                        value[struct_key] = (
                            self.decode_address(struct_val)
                            if struct_type == "address"
                            else HexBytes(struct_val)
                            if "bytes" in struct_type
                            else struct_val
                        )
                    converted_arguments[key] = value

                elif _type == "address":
                    converted_arguments[key] = self.decode_address(value)

                elif is_array(_type):
                    sub_type = "[".join(_type.split("[")[:-1])
                    converted_arguments[key] = (
                        [self.decode_address(v) for v in value] if sub_type == "address" else value
                    )

                elif isinstance(value, int):
                    # This allows integers to be comparable with currency-value
                    # strings, such as "1 ETH".
                    converted_arguments[key] = CurrencyValueComparable(value)

                else:
                    # No change.
                    converted_arguments[key] = value

            yield ContractLog(
                _abi=abi,
                block_hash=log.get("blockHash") or log.get("block_hash") or "",
                block_number=log.get("blockNumber") or log.get("block_number") or 0,
                contract_address=self.decode_address(log["address"]),
                event_arguments=converted_arguments,
                event_name=abi.event_name,
                log_index=log.get("logIndex") or log.get("log_index") or 0,
                transaction_hash=log.get("transactionHash") or log.get("transaction_hash") or "",
                transaction_index=(
                    log.get("transactionIndex")
                    if "transactionIndex" in log
                    else log.get("transaction_index")
                ),
            )

    def enrich_trace(self, trace: "TraceAPI", **kwargs) -> "TraceAPI":
        kwargs["trace"] = trace
        if not isinstance(trace, Trace):
            # Can only enrich `ape_ethereum.trace.Trace` (or subclass) implementations.
            return trace

        elif trace._enriched_calltree is not None:
            # Already enriched.
            return trace

        if sender := trace.transaction.get("from"):
            kwargs["sender"] = sender

        # Get the un-enriched calltree.
        # NOTE: Using JSON mode so Enums are all str types.
        data = trace.get_calltree().model_dump(mode="json", by_alias=True)

        if isinstance(trace, TransactionTrace):
            return_value = trace.__dict__.get("return_value") if data.get("depth", 0) == 0 else None
            if return_value is not None:
                # Return value was discovered already.
                kwargs["return_value"] = return_value

        # Cache the result back on the trace.
        trace._enriched_calltree = self._enrich_calltree(data, **kwargs)
        return trace

    def _enrich_calltree(self, call: dict, **kwargs) -> dict:
        if "contract_id" in call:
            # Already enriched.
            return call

        if self._test_runner and self._test_runner.gas_tracker.enabled:
            default_symbol_for_tokens = not self._test_runner.gas_tracker.enabled
        else:
            default_symbol_for_tokens = True

        kwargs["use_symbol_for_tokens"] = kwargs.get(
            "use_symbol_for_tokens", default_symbol_for_tokens
        )

        # Handle if for some reason this is still an Enum.
        call_type = call.get("call_type", "")
        if call_type and not isinstance(call_type, str):
            call["call_type"] = call_type = call_type.value

        is_create = "CREATE" in call_type

        # Enrich sub-calls first.
        if subcalls := call.get("calls"):
            call["calls"] = [self._enrich_calltree(c, **kwargs) for c in subcalls]

        # Figure out the contract.
        address: AddressType = call.pop("address", "")
        try:
            call["contract_id"] = address = kwargs["contract_address"] = self.decode_address(
                address
            )
        except Exception:
            # Tx was made with a weird address.
            call["contract_id"] = address

        if calldata := call.get("calldata"):
            calldata_bytes = HexBytes(calldata)
            call["method_id"] = to_hex(calldata_bytes[:4])
            call["calldata"] = calldata if is_create else to_hex(calldata_bytes[4:])

        else:
            call["method_id"] = "0x"

        try:
            address_int = int(address, 16)
        except Exception:
            pass
        else:
            # Collapse pre-compile address calls
            if 1 <= address_int <= 9:
                return (
                    call["calls"][0]
                    if len(call.get("calls", [])) == 1
                    else {"contract_id": f"{address_int}", "calls": call["calls"]}
                )

        depth = call.get("depth", 0)
        if depth == 0 and address in self.account_manager:
            call["contract_id"] = f"__{self.fee_token_symbol}_transfer__"
        else:
            call["contract_id"] = self._enrich_contract_id(call["contract_id"], **kwargs)

        if not (contract_type := self._get_contract_type_for_enrichment(address, **kwargs)):
            # Without a contract type, we can enrich no further.
            return call

        kwargs["contract_type"] = contract_type
        if events := call.get("events"):
            call["events"] = self._enrich_trace_events(events, address=address, **kwargs)

        method_abi: Optional[Union[MethodABI, ConstructorABI]] = None
        if is_create:
            method_abi = contract_type.constructor
            name = "__new__"

        elif call["method_id"] != "0x":
            method_id_bytes = HexBytes(call["method_id"])

            # perf: use try/except instead of __contains__ check.
            try:
                method_abi = contract_type.methods[method_id_bytes]
            except KeyError:
                name = call["method_id"]
            else:
                if isinstance(method_abi, MethodABI):
                    # Check if method name duplicated. If that is the case, use selector.
                    times = len([x for x in contract_type.methods if x.name == method_abi.name])
                    name = (method_abi.name if times == 1 else method_abi.selector) or call[
                        "method_id"
                    ]
                    call = self._enrich_calldata(call, method_abi, **kwargs)
                else:
                    name = call.get("method_id") or "0x"
        else:
            name = call.get("method_id") or "0x"

        call["method_id"] = name

        if method_abi:
            call = self._enrich_calldata(call, method_abi, **kwargs)

            if kwargs.get("return_value"):
                # Return value was separately enriched.
                call["returndata"] = kwargs["return_value"]
            elif isinstance(method_abi, MethodABI):
                call = self._enrich_returndata(call, method_abi, **kwargs)
            else:
                # For constructors, don't include outputs, as it is likely a large amount of bytes.
                call["returndata"] = None

        elif "revert_message" not in call:
            # Method not found but perhaps we still know the error.
            call = self._enrich_revert_message(call)

        return call

    def _enrich_contract_id(self, address: AddressType, **kwargs) -> str:
        # Defensively pop "contract_type" key from kwargs. `_get_contract_type_for_enrichment` will
        # preferentially return a `contract_type` from kwargs without checking the contract cache.
        # The contract_type may not match the contract address being enriched if this method was
        # previously called for a different contract.
        kwargs.pop("contract_type", None)

        if address and address == kwargs.get("sender"):
            return "tx.origin"

        elif address == ZERO_ADDRESS:
            return "ZERO_ADDRESS"

        elif not (contract_type := self._get_contract_type_for_enrichment(address, **kwargs)):
            # Without a contract type, we can enrich no further.
            return address

        kwargs["contract_type"] = contract_type
        if kwargs.get("use_symbol_for_tokens") and "symbol" in contract_type.view_methods:
            # Use token symbol as name
            contract = self.chain_manager.contracts.instance_at(address)

            try:
                symbol = contract.symbol(skip_trace=True)
            except ApeException:
                symbol = None

            if isinstance(symbol, str):
                return symbol.strip()

            # bytes32 symbol appears in ds-token
            if isinstance(symbol, bytes):
                try:
                    return symbol.rstrip(b"\x00").decode()
                except UnicodeDecodeError:
                    return str(symbol)

        name = contract_type.name.strip() if contract_type.name else None
        return name or address

    def _enrich_calldata(
        self,
        call: dict,
        method_abi: Union[MethodABI, ConstructorABI],
        **kwargs,
    ) -> dict:
        calldata = call["calldata"]
        if isinstance(calldata, str):
            calldata_arg = to_bytes(hexstr=calldata)
        elif isinstance(calldata, bytes):
            calldata_arg = calldata
        else:
            # Already enriched.
            return call

        contract_type = kwargs["contract_type"]
        if call.get("call_type") and "CREATE" in call.get("call_type", ""):
            # Strip off bytecode
            bytecode = (
                contract_type.deployment_bytecode.to_bytes()
                if contract_type.deployment_bytecode
                else b""
            )
            # TODO: Handle Solidity Metadata (delegate to Compilers again?)
            calldata_arg = HexBytes(calldata_arg.split(bytecode)[-1])

        try:
            call["calldata"] = self.decode_calldata(method_abi, calldata_arg)
        except DecodingError:
            call["calldata"] = ["<?>" for _ in method_abi.inputs]
        else:
            call["calldata"] = self._enrich_calldata_dict(call["calldata"], **kwargs)

        return call

    def _enrich_calldata_dict(self, calldata: dict, **kwargs) -> dict:
        return {k: self._enrich_value(v, **kwargs) for k, v in calldata.items()}

    def _enrich_returndata(self, call: dict, method_abi: MethodABI, **kwargs) -> dict:
        if "CREATE" in call.get("call_type", ""):
            call["returndata"] = ""
            return call

        elif "revert_message" in call:
            # Already enriched, in a sense..
            return call

        default_return_value = "<?>"
        returndata = call.get("returndata", "")
        is_hexstr = isinstance(returndata, str) and is_0x_prefixed(returndata)

        # Check if return is only a revert string.
        call = self._enrich_revert_message(call)
        if "revert_message" in call:
            return call

        elif is_hexstr:
            return_value_bytes = HexBytes(returndata)

            # Check if custom-error.
            if "trace" in kwargs and "contract_address" in kwargs:
                address = kwargs["contract_address"]
                try:
                    instance = self.decode_custom_error(return_value_bytes, address, **kwargs)
                except NotImplementedError:
                    pass
                else:
                    if instance is not None:
                        call["revert_message"] = repr(instance)
                        return call

        elif is_hexstr or isinstance(returndata, (int, bytes)):
            return_value_bytes = HexBytes(returndata)
        else:
            return_value_bytes = None

        if return_value_bytes is None:
            values = tuple([default_return_value for _ in method_abi.outputs])

        else:
            return_values = None
            try:
                return_values = (
                    self.decode_returndata(method_abi, return_value_bytes)
                    if not call.get("failed")
                    else None
                )
            except DecodingError:
                if return_value_bytes == HexBytes("0x"):
                    # Empty result, but it failed decoding because of its length.
                    return_values = ("",)

            # Cache un-enriched return_value in trace.
            call["unenriched_return_values"] = return_values

            values = (
                tuple([default_return_value for _ in method_abi.outputs])
                if return_values is None
                else tuple([self._enrich_value(v, **kwargs) for v in return_values or ()])
            )

        output_val = values[0] if len(values) == 1 else values
        if (
            isinstance(output_val, str)
            and is_0x_prefixed(output_val)
            and "." not in output_val
            and not int(output_val, 16)
        ):
            output_val = ""

        call["returndata"] = output_val
        return call

    def _enrich_trace_events(
        self,
        events: list[dict],
        address: Optional[AddressType] = None,
        **kwargs,
    ) -> list[dict]:
        return [self._enrich_trace_event(e, address=address, **kwargs) for e in events]

    def _enrich_trace_event(
        self,
        event: dict,
        address: Optional[AddressType] = None,
        **kwargs,
    ) -> dict:
        if "topics" not in event or len(event["topics"]) < 1:
            # Already enriched or wrong.
            return event

        elif not address:
            address = event.get("address")
            if not address:
                # Cannot enrich further w/o an address.
                return event

        if not (contract_type := self._get_contract_type_for_enrichment(address, **kwargs)):
            # Without a contract type, we can enrich no further.
            return event

        kwargs["contract_type"] = contract_type

        # The selector is always the first topic.
        selector = event["topics"][0]
        if not isinstance(selector, str):
            selector = to_hex(selector)

        if selector not in contract_type.identifier_lookup:
            # Unable to enrich using this contract type.
            # Selector unknown.
            return event

        abi = contract_type.identifier_lookup[selector]
        assert isinstance(abi, EventABI)  # For mypy.
        log_data = {
            "topics": event["topics"],
            "data": event["data"],
            "address": address,
        }

        try:
            contract_logs = [log for log in self.decode_logs([log_data], abi)]
        except Exception as err:
            logger.debug(f"Failed decoding logs from trace data: {err}")
            return event

        if not contract_logs:
            # Not sure if this is a likely condition.
            return event

        # Enrich the event-node data using the Ape ContractLog object.
        log: ContractLog = contract_logs[0]
        calldata = self._enrich_calldata_dict(log.event_arguments)
        return {"name": log.event_name, "calldata": calldata}

    def _enrich_revert_message(self, call: dict) -> dict:
        returndata = call.get("returndata", "")
        is_hexstr = isinstance(returndata, str) and is_0x_prefixed(returndata)
        if is_hexstr and returndata.startswith(_REVERT_PREFIX):
            # The returndata is the revert-str.
            decoded_result = decode(("string",), HexBytes(returndata)[4:])
            call["revert_message"] = decoded_result[0] if len(decoded_result) == 1 else ""

        return call

    def _get_contract_type_for_enrichment(
        self, address: AddressType, **kwargs
    ) -> Optional["ContractType"]:
        if not (contract_type := kwargs.get("contract_type")):
            try:
                contract_type = self.chain_manager.contracts.get(address)
            except Exception as err:
                logger.debug(f"Error getting contract type during event enrichment: {err}")

        return contract_type

    def get_python_types(self, abi_type: ABIType) -> Union[type, Sequence]:
        return self._python_type_for_abi_type(abi_type)

    def decode_custom_error(
        self,
        data: HexBytes,
        address: AddressType,
        **kwargs,
    ) -> Optional[CustomError]:
        # Use an instance (required for proper error caching).
        try:
            contract = self.chain_manager.contracts.instance_at(address)
        except Exception:
            return None

        selector = data[:4]
        input_data = data[4:]
        if selector in contract.contract_type.errors:
            abi = contract.contract_type.errors[selector]
            error_cls = contract.get_error_by_signature(abi.signature)
            inputs = self.decode_calldata(abi, input_data)
            kwargs["contract_address"] = address
            error_kwargs = {
                k: v
                for k, v in kwargs.items()
                if k in ("trace", "txn", "contract_address", "source_traceback")
            }
            return error_cls(abi, inputs, **error_kwargs)

        # ABI not found. Try looking at the "last" contract.
        if not (tx := kwargs.get("txn")) or not self.network_manager.active_provider:
            return None

        try:
            tx_hash = tx.txn_hash
        except SignatureError:
            return None

        try:
            trace = kwargs.get("trace") or self.provider.get_transaction_trace(tx_hash)
        except NotImplementedError:
            return None

        try:
            if not (last_addr := next(trace.get_addresses_used(reverse=True), None)):
                return None
        except Exception:
            # When unable to get trace-frames properly, such as eth-tester.
            return None

        if last_addr == address:
            # Avoid checking same address twice.
            return None

        try:
            if cerr := self.decode_custom_error(data, last_addr, **kwargs):
                return cerr

        except NotImplementedError:
            return None

        # error never found.
        return None

    def get_deployment_address(self, address: AddressType, nonce: int) -> AddressType:
        """
        Calculate the deployment address of a contract before it is deployed.
        This is useful if the address is an argument to another contract's deployment
        and you have not yet deployed the first contract yet.
        """
        sender_bytes = to_bytes(hexstr=address)
        encoded = rlp.encode([sender_bytes, nonce])
        address_bytes = keccak(encoded)[12:]
        return self.decode_address(address_bytes)


def parse_type(type_: dict[str, Any]) -> Union[str, tuple, list]:
    if "tuple" not in type_["type"]:
        return type_["type"]

    result = tuple([parse_type(c) for c in type_["components"]])
    return [result] if is_array(type_["type"]) else result


def _correct_key(key: str, data: dict, alt_keys: tuple[str, ...]) -> dict:
    if key in data:
        return data

    for possible_key in alt_keys:
        if possible_key in data:
            new_data = data.copy()
            new_data[key] = new_data.pop(possible_key)
            return new_data

    return data



import json
import os
import shutil
from pathlib import Path

import pytest
from ethpm_types import PackageManifest
from pydantic import ValidationError

import ape
from ape.exceptions import ProjectError
from ape.managers.project import Dependency, LocalProject, PackagesCache, Project, ProjectManager
from ape.utils import create_tempdir
from ape_pm.dependency import GithubDependency, LocalDependency, NpmDependency, PythonDependency
from tests.conftest import skip_if_plugin_installed


@pytest.fixture
def oz_dependencies_config():
    def _create_oz_dependency(version: str) -> dict:
        return {
            "name": "openzeppelin",
            "version": version,
            "github": "OpenZeppelin/openzeppelin-contracts",
        }

    return {"dependencies": [_create_oz_dependency("3.1.0"), _create_oz_dependency("4.4.2")]}


@pytest.fixture
def project_with_downloaded_dependencies(project, oz_dependencies_config):
    # The path to source-test data.
    manifests_directory = Path(__file__).parent / "data" / "manifests"
    oz_manifests = manifests_directory / "openzeppelin"

    # Copy test-data in the DATA_FOLDER, as if these were already fetched.
    base = project.dependencies.packages_cache
    package_id = "OpenZeppelin_openzeppelin-contracts"
    oz_manifests_dest = base.manifests_folder / package_id
    oz_manifests_dest.mkdir(exist_ok=True, parents=True)

    for version in ("3.1.0", "4.4.2"):
        manifest_source = oz_manifests / version / "openzeppelin.json"
        manifest_dest = oz_manifests_dest / f"{version.replace('.', '_')}.json"
        manifest_dest.unlink(missing_ok=True)
        manifest_dest.write_text(manifest_source.read_text(), encoding="utf8")

        # Also, copy in the API data
        api_dest = base.api_folder / package_id / f"{version.replace('.', '_')}.json"
        api_dest.unlink(missing_ok=True)
        cfg = [x for x in oz_dependencies_config["dependencies"] if x["version"] == version][0]
        api_dest.parent.mkdir(exist_ok=True, parents=True)
        api_dest.write_text(json.dumps(cfg), encoding="utf8")

    with project.temp_config(**oz_dependencies_config):
        yield project


def test_repr(project):
    assert isinstance(project, LocalProject), "Setup failed - expecting local project"
    actual = repr(project.dependencies)
    path = str(project.path).replace(str(Path.home()), "$HOME")
    expected = f"<DependencyManager project={path}>"
    assert actual == expected


def test_repr_manifest_project():
    """
    Tests against assuming the dependency manager is related
    to a local project (could be from a manifest-project).
    """
    manifest = PackageManifest()
    project = Project.from_manifest(manifest, config_override={"name": "testname123"})
    dm = project.dependencies
    actual = repr(dm)
    expected = "<DependencyManager project=testname123>"
    assert actual == expected


def test_len(project):
    actual = len(project.dependencies)
    expected = len(project.config.dependencies)
    assert actual == expected


@pytest.mark.parametrize("ref", ("main", "v1.0.0", "1.0.0"))
def test_decode_dependency_using_reference(ref, recwarn, project):
    dependency_config = {
        "github": "apeworx/testfoobartest",
        "name": "foobar",
        "ref": ref,
    }
    dependency = project.dependencies.decode_dependency(**dependency_config)
    assert dependency.version is None
    assert dependency.ref == ref
    assert str(dependency.uri) == f"https://github.com/apeworx/testfoobartest/tree/{ref}"

    # Tests against bug where we tried creating version objects out of branch names,
    # thus causing warnings to show in `ape test` runs.
    assert DeprecationWarning not in [w.category for w in recwarn.list]


def test_dependency_with_multiple_versions(project_with_downloaded_dependencies):
    """
    Testing the case where we have OpenZeppelin installed multiple times
    with different versions.
    """
    name = "openzeppelin"
    dm = project_with_downloaded_dependencies.dependencies

    # We can get both projects at once! One for each version.
    oz_310 = dm.get_dependency(name, "3.1.0", allow_install=False)
    oz_442 = dm.get_dependency(name, "4.4.2", allow_install=False)
    base_uri = "https://github.com/OpenZeppelin/openzeppelin-contracts/releases/tag"

    assert oz_310.version == "3.1.0"
    assert oz_310.name == name
    assert str(oz_310.uri) == f"{base_uri}/v3.1.0"
    assert oz_442.version == "4.4.2"
    assert oz_442.name == name
    assert str(oz_442.uri) == f"{base_uri}/v4.4.2"


def test_decode_dependency_with_config_override(project):
    with project.isolate_in_tempdir() as tmp_project:
        settings = {".json": {"evm_version": "paris"}}
        path = "__test_path__"
        base_path = tmp_project.path / path
        contracts_path = base_path / "contracts"
        contracts_path.mkdir(parents=True, exist_ok=True)
        (contracts_path / "contract.json").write_text('{"abi": []}', encoding="utf8")

        data = {"name": "FooBar", "local": path, "config_override": settings}
        dependency = tmp_project.dependencies.decode_dependency(**data)
        assert dependency.config_override == settings


def test_uri_map(project_with_dependency_config):
    actual = project_with_dependency_config.dependencies.uri_map
    here = Path(__file__).parent
    # Wrap in Path to handle Windows.
    expected = Path(f"file://{here}/data/projects/LongContractsFolder")
    assert "testdependency" in actual
    assert Path(str(actual["testdependency"])) == expected


def test_get_dependency_by_package_id(project_with_downloaded_dependencies):
    dm = project_with_downloaded_dependencies.dependencies
    actual = dm.get_dependency("OpenZeppelin/openzeppelin-contracts", "4.4.2")
    assert actual.name == "openzeppelin"
    assert actual.version == "4.4.2"
    assert actual.package_id == "OpenZeppelin/openzeppelin-contracts"


def test_get_dependency_by_name(project_with_downloaded_dependencies):
    dm = project_with_downloaded_dependencies.dependencies
    actual = dm.get_dependency("openzeppelin", "4.4.2")
    assert actual.name == "openzeppelin"
    assert actual.version == "4.4.2"
    assert actual.package_id == "OpenZeppelin/openzeppelin-contracts"


def test_get_dependency_returns_latest_updates(project, project_with_contracts):
    """
    Integration test showing how to utilize the changes to dependencies from the config.
    """
    cfg = [{"name": "dep", "version": "10.1.10", "local": f"{project_with_contracts.path}"}]
    with project.temp_config(dependencies=cfg):
        project.dependencies.install()

        dep = project.dependencies.get_dependency("dep", "10.1.10")
        assert dep.package_id == f"{project_with_contracts.path}"
        assert dep.api.local == project_with_contracts.path

        # Changing the version in the config and force re-installing should change the package data.
        project.config["dependencies"][0]["local"] = f"{Path(__file__).parent}"
        project.dependencies.install(use_cache=False)

        # Show `.get_dependency()` will return the updated version when given name alone.
        dep = project.dependencies.get_dependency("dep", "10.1.10")
        assert dep.api.local == Path(__file__).parent


def test_get_versions_using_package_id(project_with_downloaded_dependencies):
    dm = project_with_downloaded_dependencies.dependencies
    package_id = "OpenZeppelin/openzeppelin-contracts"
    actual = list(dm.get_versions(package_id))
    assert len(actual) == 2


def test_get_versions_using_name(project_with_downloaded_dependencies):
    dm = project_with_downloaded_dependencies.dependencies
    name = "openzeppelin"
    actual = list(dm.get_versions(name))
    assert len(actual) == 2


def test_getitem_and_contains_and_get(project_with_downloaded_dependencies):
    dm = project_with_downloaded_dependencies.dependencies
    name = "openzeppelin"
    versions = dm[name]
    assert "3.1.0" in versions
    assert "v3.1.0" in versions  # Also allows v-prefix.
    assert (
        versions["3.1.0"] == versions["v3.1.0"] == versions.get("3.1.0") == versions.get("v3.1.0")
    )
    assert isinstance(versions["3.1.0"], LocalProject)


def test_add(project):
    with project.isolate_in_tempdir() as tmp_project:
        contracts_path = tmp_project.path / "src"
        contracts_path.mkdir(exist_ok=True, parents=True)
        (contracts_path / "contract.json").write_text('{"abi": []}', encoding="utf8")
        data = {"name": "FooBar", "local": f"{tmp_project.path}"}

        dependency = project.dependencies.add(data)
        assert isinstance(dependency, Dependency)
        assert dependency.name == "foobar"
        assert dependency.version == "local"

        # After adding, we should be able to "get" it.
        dep2 = project.dependencies.get_dependency(
            dependency.package_id, dependency.version, allow_install=False
        )
        assert dep2 == dependency

        # Attempt to add again.
        dep3 = project.dependencies.add(data)
        assert dep3 == dependency


def test_add_dependency_with_dependencies(project, with_dependencies_project_path):
    dm = project.dependencies
    data = {"name": "wdep", "local": with_dependencies_project_path}
    actual = dm.add(data)
    assert actual.name == "wdep"
    assert actual.version == "local"


def test_get_project_dependencies(project, ape_caplog):
    installed_package = {"name": "web3", "site_package": "web3"}
    not_installed_package = {
        "name": "apethisisnotarealpackageape",
        "site_package": "apethisisnotarealpackageape",
    }
    with project.temp_config(dependencies=[installed_package, not_installed_package]):
        dm = project.dependencies
        actual = list(dm.get_project_dependencies())
        logs = ape_caplog.head
        assert len(actual) == 2
        assert actual[0].name == "web3"
        assert actual[0].installed
        assert actual[1].name == "apethisisnotarealpackageape"
        assert not actual[1].installed
        assert "Dependency 'apethisisnotarealpackageape' not installed." in logs


def test_install(project, mocker):
    with project.isolate_in_tempdir() as tmp_project:
        contracts_path = tmp_project.path / "src"
        contracts_path.mkdir(exist_ok=True, parents=True)
        (contracts_path / "contract.json").write_text('{"abi": []}', encoding="utf8")
        data = {"name": "FooBar", "local": f"{tmp_project.path}"}
        get_spec_spy = mocker.spy(tmp_project.dependencies, "get_project_dependencies")
        install_dep_spy = mocker.spy(tmp_project.dependencies, "install_dependency")

        # Show can install from DependencyManager.
        dependency = tmp_project.dependencies.install(**data)
        assert isinstance(dependency, Dependency)
        # NOTE: Here, we are mostly testing that `use_cache=False` was not passed.
        assert install_dep_spy.call_count == 1

        # Show can install from Dependency.
        project = dependency.install()
        assert isinstance(project, ProjectManager)

        # Delete project path (but not manifest) and install again.
        shutil.rmtree(dependency.project_path)
        installation = dependency._installation
        dependency._installation = None
        project = dependency.install()
        dependency._installation = installation
        assert isinstance(project, ProjectManager)
        assert dependency.project_path.is_dir()  # Was re-created from manifest sources.

        # Force install and prove it actually does the re-installation.
        tmp_project.dependencies.install(use_cache=False)
        get_spec_spy.assert_called_once_with(use_cache=False, allow_install=True, recurse=True)


def test_install_dependencies_of_dependencies(project, with_dependencies_project_path):
    dm = project.dependencies
    actual = dm.install(local=with_dependencies_project_path, name="wdep")
    assert actual.name == "wdep"

    # Ensure dependencies of dependencies also installed.
    dep_with_deps = actual.project.dependencies["containing-sub-dependencies"]["local"]
    dep_of_dep = dep_with_deps.dependencies.get_dependency("sub-dependency", "local")
    assert dep_of_dep.installed


def test_install_already_installed(mocker, project, with_dependencies_project_path):
    """
    Some dependencies never produce sources because of default compiler extension behavior
    (example: a16z_erc4626-tests repo).
    """
    dm = project.dependencies
    dep = dm.install(local=with_dependencies_project_path, name="wdep")

    # Set up a spy on the fetch API, which can be costly and require internet.
    real_api = dep.api
    mock_api = mocker.MagicMock()
    dep.api = mock_api

    # Remove in-memory cache.
    dep._installation = None

    # Install again.
    with pytest.raises(ProjectError):
        dep.install()

    dep.api = real_api

    # Ensure it doesn't need to re-fetch
    assert mock_api.call_count == 0


@pytest.mark.parametrize("name", ("openzeppelin", "OpenZeppelin/openzeppelin-contracts"))
def test_uninstall(name, project_with_downloaded_dependencies):
    version = "4.4.2"
    dm = project_with_downloaded_dependencies.dependencies
    dependency = dm.get_dependency(name, version)
    dependency.uninstall()
    assert not any(
        (d.name == name or d.package_id == name) and d.version == version for d in dm.all
    )


def test_unpack(project_with_downloaded_dependencies):
    dm = project_with_downloaded_dependencies.dependencies
    dep = dm.get_dependency("openzeppelin", "4.4.2")
    with create_tempdir() as tempdir:
        created = list(dep.unpack(tempdir))
        assert len(created) == 1
        assert created[0].package_id == "OpenZeppelin/openzeppelin-contracts"

        files = [x.name for x in (tempdir / "openzeppelin" / "4.4.2" / "contracts").iterdir()]
        assert "token" in files
        assert "access" in files


def test_unpack_dependencies_of_dependencies(project, with_dependencies_project_path):
    dep = project.dependencies.install(local=with_dependencies_project_path, name="wdep")
    with create_tempdir() as tempdir:
        list(dep.unpack(tempdir))
        subdirs = [x.name for x in tempdir.iterdir() if x.is_dir()]
        assert "sub-dependency" in subdirs


def test_unpack_no_contracts_folder(project, with_dependencies_project_path):
    dep = project.dependencies.install(local=with_dependencies_project_path, name="wdep")
    with create_tempdir() as tempdir:
        list(dep.unpack(tempdir))
        subdirs = [x.name for x in tempdir.iterdir() if x.is_dir()]
        assert "empty-dependency" in subdirs


def test_all(project):
    # Set up an installed dependency.
    packages_cache = project.dependencies.packages_cache
    corrupt_file = packages_cache.api_folder / "dep123" / "0_4_0.json"
    corrupt_file.parent.mkdir(exist_ok=True, parents=True)
    corrupt_api = {"name": "dep123", "local": "."}
    corrupt_file.write_text(json.dumps(corrupt_api), encoding="utf8")

    # Iterate through the installed list.
    deps = [x for x in project.dependencies.all if x.name == "dep123"]

    # Show it is there.
    assert len(deps) >= 1
    assert deps[0].name == "dep123"
    assert not deps[0].installed


def test_all_when_contains_corrupt_dependency(project):
    packages_cache = project.dependencies.packages_cache
    name = "anothermissingpackagefromanapietest"
    corrupt_file = packages_cache.api_folder / name / "0_1_0.json"
    corrupt_file.parent.mkdir(exist_ok=True, parents=True)
    corrupt_api = {"name": name, "site_package": name}
    corrupt_file.write_text(json.dumps(corrupt_api), encoding="utf8")
    deps = [x for x in project.dependencies.all if x.name == name]
    assert len(deps) > -1
    assert not deps[0].installed


class TestPackagesCache:
    @pytest.fixture
    def cache(self):
        return PackagesCache()

    def test_root(self, cache, data_folder):
        actual = cache.root
        expected = data_folder / "packages"
        assert actual == expected

    def test_api_folder(self, cache, data_folder):
        actual = cache.api_folder
        expected = data_folder / "packages" / "api"
        assert actual == expected

    def test_get_api_path(self, cache, data_folder):
        package_id = "this/is/my_package-ID"
        version = "version12/5.54"
        actual = cache.get_api_path(package_id, version)
        expected = (
            data_folder / "packages" / "api" / "this_is_my_package-ID" / "version12_5_54.json"
        )
        assert actual == expected

    def test_cache_api(self, cache):
        dep = LocalDependency(name="depabc", local=Path("depabc"))
        path = cache.cache_api(dep)
        assert path.is_file()
        assert path == cache.get_api_path("depabc", "local")
        actual = json.loads(path.read_text())
        assert actual == {"name": "depabc", "local": "depabc"}

    def test_cache_api_when_v_prefix_exists(self, cache):
        # First, cache the v-prefix
        dep = LocalDependency(name="depabc", local=Path("depabc"), version="v1.0.0")
        path0 = cache.cache_api(dep)
        assert path0.is_file()

        # Now, try to cache again w/o v prefix
        dep.version = "1.0.0"
        path1 = cache.cache_api(dep)
        assert path1 == path0

    def test_cache_api_when_non_v_prefix_exists(self, cache):
        # First, cache the non-v-prefix
        dep = LocalDependency(name="depabc", local=Path("depabc"), version="1.0.0")
        path0 = cache.cache_api(dep)
        assert path0.is_file()

        # Now, try to cache again with the v prefix
        dep.version = "v1.0.0"
        path1 = cache.cache_api(dep)
        assert path1 == path0

    def test_get_manifest_path(self, cache, data_folder):
        package_id = "this/is/my_package-ID"
        version = "version12/5.54"
        actual = cache.get_manifest_path(package_id, version)
        expected = (
            data_folder / "packages" / "manifests" / "this_is_my_package-ID" / "version12_5_54.json"
        )
        assert actual == expected

    def test_get_manifest_path_v_prefix_exists(self, cache, data_folder):
        file = data_folder / "packages" / "manifests" / "manifest-pkg-test-1" / "v1_0_0.json"
        file.parent.mkdir(parents=True, exist_ok=True)
        file.touch()

        # Requesting w/o v-prefix. Should still work.
        path = cache.get_manifest_path("manifest-pkg-test-1", "1.0.0")
        assert path == file

    def test_get_manifest_path_non_v_prefix_exists(self, cache, data_folder):
        file = data_folder / "packages" / "manifests" / "manifest-pkg-test-2" / "1_0_0.json"
        file.parent.mkdir(parents=True, exist_ok=True)
        file.touch()

        # Requesting w/o v-prefix. Should still work.
        path = cache.get_manifest_path("manifest-pkg-test-2", "v1.0.0")
        assert path == file

    def test_get_project_path(self, cache, data_folder):
        path = data_folder / "packages" / "projects" / "project-test-1" / "local"
        actual = cache.get_project_path("project-test-1", "local")
        assert actual == path

    def test_get_project_path_missing_v_prefix(self, cache, data_folder):
        path = data_folder / "packages" / "projects" / "project-test-1" / "v1_0_0"
        path.mkdir(parents=True, exist_ok=True)
        # Missing v-prefix on request.
        actual = cache.get_project_path("project-test-1", "1.0.0")
        assert actual == path

    def test_get_project_path_unneeded_v_prefix(self, cache, data_folder):
        path = data_folder / "packages" / "projects" / "project-test-2" / "1_0_0"
        path.mkdir(parents=True, exist_ok=True)
        # Unneeded v-prefix on request.
        actual = cache.get_project_path("project-test-2", "v1.0.0")
        assert actual == path

    def test_isolate_cache_changes(self, cache):
        dep = LocalDependency(name="isotestdep", local=Path("isotestdep"), version="v1.0.0")
        with cache.isolate_changes():
            path = cache.cache_api(dep)
            assert path.is_file()

        assert not path.is_file()


class TestLocalDependency:
    NAME = "testlocaldep"
    VERSION = "1.0.0"
    PATH = Path.cwd()

    @pytest.fixture
    def dependency(self):
        return LocalDependency(local=self.PATH, name=self.NAME, version=self.VERSION)

    @property
    def clean_path(self) -> str:
        return str(self.PATH).replace(str(Path.home()), "$HOME")

    def test_repr(self, dependency):
        actual = repr(dependency)
        expected = f"<LocalDependency local={self.clean_path}, version={self.VERSION}>"
        assert actual == expected

    def test_name(self, dependency):
        assert dependency.name == self.NAME

    def test_version(self, dependency):
        assert dependency.version == self.VERSION

    def test_uri(self, dependency):
        assert dependency.uri == self.PATH.as_uri()

    def test_local_expands_user(self):
        path = "~/path/to/dep"
        dependency = LocalDependency(local=path, name=self.NAME, version=self.VERSION)
        assert "~" not in f"{dependency.local}"


class TestNpmDependency:
    NAME = "@gnosis.pm"
    PACKAGE = "safe-singleton-factory"
    VERSION = "1.0.0"

    @pytest.fixture
    def project_with_npm_dependency(self):
        with create_tempdir() as temp_path:
            yield ape.Project(temp_path)

    @pytest.fixture(params=("local", "global"))
    def node_modules_path(self, project_with_npm_dependency, request, mock_home_directory):
        pm = project_with_npm_dependency
        base = pm.path if request.param == "local" else mock_home_directory
        package_folder = base / "node_modules" / self.NAME / self.PACKAGE
        contracts_folder = package_folder / "contracts"
        contracts_folder.mkdir(parents=True)
        package_json = package_folder / "package.json"
        package_json.write_text(f'{{"version": "{self.VERSION}"}}', encoding="utf8")
        file = contracts_folder / "contract.json"
        source_content = '{"abi": []}'
        file.write_text(source_content, encoding="utf8")
        return base

    def test_fetch(self, node_modules_path, project_with_npm_dependency):
        pm = project_with_npm_dependency
        dependency = NpmDependency(
            name=self.PACKAGE,
            npm=f"{self.NAME}/{self.PACKAGE}",
            version=self.VERSION,
            project=pm,
        )
        dependency.fetch(pm.path / "foo")
        assert len([x for x in (pm.path / "foo").iterdir()]) > 0


class TestGitHubDependency:
    @pytest.fixture
    def mock_client(self, mocker):
        return mocker.MagicMock()

    def test_ref_or_version_is_required(self):
        expected = r"GitHub dependency must have either ref or version specified"
        with pytest.raises(ValidationError, match=expected):
            _ = GithubDependency(name="foo", github="asdf")

    def test_name_from_github(self):
        """
        When not given a name, it is derived from the github suffix.
        """
        dependency = GithubDependency(  # type: ignore
            github="ApeWorX/ApeNotAThing", version="3.0.0"
        )
        assert dependency.name == "apenotathing"

    def test_fetch_given_version(self, mock_client):
        dependency = GithubDependency(
            github="ApeWorX/ApeNotAThing", version="3.0.0", name="apetestdep"
        )
        dependency._github_client = mock_client
        with create_tempdir() as path:
            dependency.fetch(path)

        mock_client.download_package.assert_called_once_with(
            "ApeWorX", "ApeNotAThing", "3.0.0", path
        )

    def test_fetch_missing_v_prefix(self, mock_client):
        """
        Show if the version expects a v-prefix but you don't
        provide one that it still works.
        """
        dependency = GithubDependency(
            github="ApeWorX/ApeNotAThing", version="3.0.0", name="apetestdep"
        )
        dependency._github_client = mock_client

        # Simulate only v-prefix succeeding from GH
        def only_want_v(n0, n1, vers, pth):
            if not vers.startswith("v"):
                raise ValueError("nope")

        mock_client.download_package.side_effect = only_want_v

        with create_tempdir() as path:
            dependency.fetch(path)

        calls = mock_client.download_package.call_args_list
        assert mock_client.download_package.call_count == 2
        # Show it first tried w/o v
        assert calls[0][0] == ("ApeWorX", "ApeNotAThing", "3.0.0", path)
        # The second call has the v!
        assert calls[1][0] == ("ApeWorX", "ApeNotAThing", "v3.0.0", path)

    def test_fetch_unneeded_v_prefix(self, mock_client):
        """
        Show if the version expects not to have a v-prefix but you
        provide one that it still works.
        """
        dependency = GithubDependency(
            github="ApeWorX/ApeNotAThing", version="v3.0.0", name="apetestdep"
        )
        dependency._github_client = mock_client

        # Simulate only non-v-prefix succeeding from GH
        def only_want_non_v(n0, n1, vers, pth):
            if vers.startswith("v"):
                raise ValueError("nope")

        mock_client.download_package.side_effect = only_want_non_v

        with create_tempdir() as path:
            dependency.fetch(path)

        calls = mock_client.download_package.call_args_list
        assert mock_client.download_package.call_count == 2
        # Show it first tried with the v
        assert calls[0][0] == ("ApeWorX", "ApeNotAThing", "v3.0.0", path)
        # The second call does not have the v!
        assert calls[1][0] == ("ApeWorX", "ApeNotAThing", "3.0.0", path)

    def test_fetch_given_version_when_expects_reference(self, mock_client):
        """
        Show that if a user configures `version:`, but version fails, it
        tries `ref:` instead as a backup.
        """
        dependency = GithubDependency(
            github="ApeWorX/ApeNotAThing", version="v3.0.0", name="apetestdep"
        )
        dependency._github_client = mock_client
        # Simulate no versions ever found on GH Api.
        mock_client.download_package.side_effect = ValueError("nope")

        # Simulate only the non-v prefix ref working (for a fuller flow)
        def needs_non_v_prefix_ref(n0, n1, dst_path, branch, scheme):
            # NOTE: This assertion is very important!
            #  We must only give it non-existing directories.
            assert not dst_path.is_dir()

            if branch.startswith("v"):
                raise ValueError("nope")

        mock_client.clone_repo.side_effect = needs_non_v_prefix_ref

        with create_tempdir() as path:
            dependency.fetch(path)

        calls = mock_client.clone_repo.call_args_list
        assert mock_client.clone_repo.call_count == 2
        # Show it first tried with the v
        assert calls[0][0] == ("ApeWorX", "ApeNotAThing", path)
        assert calls[0][1] == {"branch": "v3.0.0", "scheme": "https"}
        # The second call does not have the v!
        assert calls[1][0] == ("ApeWorX", "ApeNotAThing", path)
        assert calls[1][1] == {"branch": "3.0.0", "scheme": "https"}

    def test_fetch_ref(self, mock_client):
        """
        When specifying ref, it does not try version API at all.
        """
        dependency = GithubDependency(github="ApeWorX/ApeNotAThing", ref="3.0.0", name="apetestdep")
        dependency._github_client = mock_client

        with create_tempdir() as path:
            dependency.fetch(path)

        assert mock_client.download_package.call_count == 0
        mock_client.clone_repo.assert_called_once_with(
            "ApeWorX",
            "ApeNotAThing",
            path,
            branch="3.0.0",
            scheme="https",
        )

    def test_fetch_existing_destination_with_read_only_files(self, mock_client):
        """
        Show it handles when the destination contains read-only files already
        """
        dependency = GithubDependency(github="ApeWorX/ApeNotAThing", ref="3.0.0", name="apetestdep")
        dependency._github_client = mock_client

        with create_tempdir() as path:
            readonly_file = path / "readme.txt"
            readonly_file.write_text("readme!")

            # NOTE: This only makes a difference on Windows. If using a UNIX system,
            #   rmtree still deletes readonly files regardless. Windows is more restrictive.
            os.chmod(readonly_file, 0o444)  # Read-only permissions

            dependency.fetch(path)
            assert not readonly_file.is_file()

    def test_ssh(self, mock_client):
        dependency = GithubDependency(
            github="ApeWorX/ApeNotAThing", ref="3.0.0", name="apetestdep", scheme="ssh"
        )
        dependency._github_client = mock_client
        with create_tempdir() as path:
            dependency.fetch(path)

        assert mock_client.clone_repo.call_args[-1]["scheme"] == "ssh"


class TestPythonDependency:
    @pytest.fixture(scope="class", params=("site_package", "python", "pypi"))
    def python_dependency(self, request):
        return PythonDependency.model_validate({request.param: "web3"})

    def test_name(self, python_dependency):
        assert python_dependency.name == "web3"

    def test_version_id(self, python_dependency):
        actual = python_dependency.version_id
        assert isinstance(actual, str)
        assert len(actual) > 0
        assert actual[0].isnumeric()
        assert "." in actual  # sep from minor / major / patch

    def test_version_id_found_from_pypi(self):
        name = "xxthisnameisnotarealpythonpackagexx"
        dependency = PythonDependency.model_validate({"site_package": name})
        dependency.__dict__["package_data"] = {"info": {"version": "4.5.6"}}
        assert dependency.version_id == "4.5.6"

    def test_version_id_not_found(self):
        name = "xxthisnameisnotarealpythonpackagexx"
        dependency = PythonDependency.model_validate({"site_package": name})
        expected = f"Dependency '{name}' not installed."
        with pytest.raises(ProjectError, match=expected):
            _ = dependency.version_id

    def test_version_id_configured(self):
        name = "xxthisnameisnotarealpythonpackagexx"
        dependency = PythonDependency.model_validate({"site_package": name, "version": "3.1.2"})
        assert dependency.version == "3.1.2"

    def test_fetch(self, python_dependency):
        with create_tempdir() as temp_dir:
            python_dependency.fetch(temp_dir)
            files = [x for x in temp_dir.iterdir()]
            assert len(files) > 0


def test_specified(project, project_with_contracts):
    cfg = [{"name": "dep", "version": "0.1.0", "local": f"{project_with_contracts}"}]
    with project.temp_config(dependencies=cfg):
        actual = [x for x in project.dependencies.specified]
        assert len(actual) == 1
        assert actual[0].name == "dep"
        assert actual[0].version == "0.1.0"


class TestDependency:
    @pytest.fixture
    def api(self):
        return LocalDependency(local=Path.cwd(), name="ooga", version="1.0.0")

    @pytest.fixture
    def dependency(self, api, project):
        return Dependency(api, project)

    def test_repr(self, dependency):
        actual = repr(dependency)
        path = str(Path.cwd()).replace(str(Path.home()), "$HOME")
        expected = f"<LocalDependency local={path}, version=1.0.0>"
        assert actual == expected

    def test_project_path(self, dependency, data_folder):
        actual = dependency.project_path
        name = dependency.api.package_id.replace("/", "_")
        expected = data_folder / "packages" / "projects" / name / "1_0_0"
        assert actual == expected

    def test_api_path(self, dependency, data_folder):
        actual = dependency.api_path
        name = dependency.api.package_id.replace("/", "_")
        expected = data_folder / "packages" / "api" / name / "1_0_0.json"
        assert actual == expected

    def test_manifest_path(self, dependency, data_folder):
        actual = dependency.manifest_path
        name = dependency.api.package_id.replace("/", "_")
        expected = data_folder / "packages" / "manifests" / name / "1_0_0.json"
        assert actual == expected

    def test_installed(self, dependency):
        dependency.uninstall()
        assert not dependency.installed
        dependency.install()
        assert dependency.installed

    def test_installed_version_id_fails(self, project):
        api = PythonDependency(
            site_package="apethisdependencyisnotinstalled",
            name="apethisdependencyisnotinstalled",
        )
        dependency = Dependency(api, project)
        assert not dependency.installed

    def test_compile(self, project):
        with create_tempdir() as path:
            api = LocalDependency(local=path, name="ooga", version="1.0.0")
            dependency = Dependency(api, project)
            contract_path = dependency.project.contracts_folder / "CCC.json"
            contract_path.parent.mkdir(exist_ok=True, parents=True)
            contract_path.write_text(
                '[{"name":"foo","type":"fallback", "stateMutability":"nonpayable"}]',
                encoding="utf8",
            )

            # Since we are adding a file mid-session, we have to refresh so
            # it's picked up. Users typically don't have to do this.
            dependency.project.refresh_sources()

            result = dependency.compile()
            assert len(result) == 1
            assert result["CCC"].name == "CCC"

    @skip_if_plugin_installed("vyper", "solidity")
    def test_compile_missing_compilers(self, project, ape_caplog):
        with create_tempdir() as path:
            api = LocalDependency(local=path, name="ooga2", version="1.1.0")
            dependency = Dependency(api, project)
            sol_path = dependency.project.contracts_folder / "Sol.sol"
            sol_path.parent.mkdir(exist_ok=True, parents=True)
            sol_path.write_text("// Sol", encoding="utf8")
            vy_path = dependency.project.contracts_folder / "Vy.vy"
            vy_path.write_text("# Vy", encoding="utf8")
            expected = (
                "Compiling dependency produced no contract types. "
                "Try installing 'ape-solidity' or 'ape-vyper'."
            )
            result = dependency.compile()
            assert len(result) == 0
            assert expected in ape_caplog.head


class TestProject:
    """
    All tests related to a dependency's project.
    """

    @pytest.fixture
    def project_from_dependency(self, project_with_dependency_config):
        dependencies = project_with_dependency_config.dependencies
        return dependencies["testdependency"]["releases/v6"]

    def test_path(self, project_from_dependency):
        assert project_from_dependency.path.is_dir()

    def test_contracts_folder(self, project_from_dependency):
        """
        The local dependency fixture uses a longer contracts folder path.
        This test ensures that the contracts folder field is honored, specifically
        In the case when it contains sub-paths.
        """
        actual = project_from_dependency.contracts_folder
        expected = project_from_dependency.path / "source" / "v0.1"
        assert actual == expected

    def test_load_contracts(self, project_with_downloaded_dependencies):
        name = "openzeppelin"
        options = project_with_downloaded_dependencies.dependencies[name]
        project = options["4.4.2"]
        # NOTE: the test data is pre-compiled because the ape-solidity plugin is required.
        actual = project.load_contracts()
        assert len(actual) > 0

    def test_load_contracts_with_config_override(self, project):
        with project.isolate_in_tempdir() as tmp_project:
            # NOTE: It is important that `contracts_folder` is present in settings
            #  for this test to test against a previous bug where we got multiple values.
            override = {"contracts_folder": "src"}
            contracts_path = tmp_project.path / "src"
            contracts_path.mkdir(exist_ok=True, parents=True)
            (contracts_path / "contract.json").write_text('{"abi": []}', encoding="utf8")

            # use_cache=False only to lessen stateful test clobbering.
            data = {
                "name": "FooBar",
                "local": f"{tmp_project.path}",
                "config_override": override,
                "use_cache": False,
            }
            tmp_project.dependencies.install(**data)
            proj = tmp_project.dependencies.get("foobar", "local")
            assert proj.config.contracts_folder == "src"

            assert proj.contracts_folder == proj.path / "src"
            assert [x.name for x in proj.sources.paths] == ["contract.json"]
            actual = proj.load_contracts()
            assert len(actual) > 0

    def test_getattr(self, project_with_downloaded_dependencies):
        """
        Access contracts using __getattr__ from the dependency's project.
        """
        name = "openzeppelin"
        oz_442 = project_with_downloaded_dependencies.dependencies.get_dependency(name, "4.4.2")
        contract = oz_442.project.AccessControl
        assert contract.contract_type.name == "AccessControl"
